import random

import pygame
from Mind import Orientation, Existence

from . import consts, tools


class MutableBool:
    def __init__(self, bool_):
        self.value = bool_

    def __bool__(self):
        return self.value

    def set(self, bool_):
        self.value = bool_


class SwitchLayer(Orientation.layer):
    def __init__(self, name, props, mapping, Map):
        super().__init__(name, props, mapping, Map)
        self.on = True

    def turn_on(self):
        self.on = True

    def turn_off(self):
        self.on = False

    def blit(self):
        if self.on:
            super().blit()


class Display:
    def __init__(self, screen):
        self.screen = screen
        self.width = self.screen.get_width()
        self.time = pygame.time.get_ticks()
        self.old_time = 0
        self.font = consts.FONTS[1]
        self.names = ["live", "cold"]
        self.rects = []
        for p, x in enumerate((500, 0, 500)):
            self.rects.append(pygame.Rect(10, 10 + p * 30, x, 20))
        self.colors = [consts.LIFE, consts.COLD, consts.FOOD]

    def set_width(self, *values):
        for p, value in enumerate(values):
            self.rects[p].width = value

    def blit(self):
        time = pygame.time.get_ticks() - self.time + self.old_time
        sec, milli = divmod(time, 1000)
        m, sec = divmod(sec, 60)
        image = self.font.render("{}:{:02d}:{:03d}".format(m, sec, milli),
                                 True, consts.RED)
        self.screen.blit(image, (self.width - image.get_width() - 20, 20))
        for r, c in zip(self.rects, self.colors):
            if r.width:
                pygame.draw.rect(self.screen, c, r)

    def stop(self):
        self.old_time += pygame.time.get_ticks() - self.time

    def start(self):
        self.time = pygame.time.get_ticks()

@Existence.logic_class
class ItemGenerator:
    pictures = {
        "bottle": pygame.image.load(consts.join(consts.ART_PATH,
                                                "bottle.png")),
        "sandwich": pygame.image.load(consts.join(consts.ART_PATH,
                                                "sandwich.png"))
    }
    probs = {  # probabilities in 1/probs[item]
        "bottle": 4000,
        "sandwich": 2000
    }
    def __init__(self):
        self.item = self.props["item"]
        self.prob = self.probs[self.item]
        self.contain = False
        self.picture2 = self.pictures[self.item]
        old_height = self.obj.height
        self.obj.width, self.obj.height = self.picture2.get_size()
        self.width = self.obj.width
        self.height = self.obj.height
        self.y += old_height - self.height

    def blit(self):
        if not self.contain and not random.randrange(self.prob):
            self.contain = True
        if self.contain and not self.props["outside"]:
            self.picture = self.picture2
        else:
            self.picture = None

    def available(self):
        return self.contain

    def take(self):
        self.contain = False
        self.picture = None
        return self.item


@Existence.logic_class
class Cactus:
    def __init__(self):
        self.picture2 = self.picture
        self.picture = None
        self.width = self.obj.width
        self.height = self.obj.height
        self.player = self.Map.clone_obj("player", "type")
        self.ground = False
        self.processed = False
        self.trace = False

    def blit(self):
        if not self.processed:
            cl_obj = self.cl_objects[:]
            self.cl_objects = []
            for obj in cl_obj:
                if "out" not in obj.opts:
                    self.cl_objects.append(obj)
            self.processed = True

        if self.props["outside"]:
            self.picture = None
        else:
            self.picture = self.picture2
            
            if (tools.distance(self.x, self.y, self.player.x, self.player.y) <
                    500):
                dx = max(-3.5, min(3.5, self.player.x - self.x))
                self.move(dx, 0)
                if self.y > self.player.y and self.ground:
                    self.fy -= 9
                    self.ground = False
    
    def left(self, x, obj):
        if obj.type == "wall":
            self.set_position(obj.obj.x - self.width, self.y)

    def real_down(self, x, obj):
        if obj.type == "wall":
            self.set_position(self.x, obj.obj.y - self.height)
            self.fy = 0
            self.ground = obj
        if (obj.type == "platform" and self.fy > 0.1):
            self.set_position(self.x, obj.obj.y - self.height)
            self.fy = 0
            self.ground = obj

    def real_up(self, x, obj):
        if obj.type == "wall":
            self.set_position(self.x, obj.obj.y + obj.obj.height)
            self.fy = 0

    def down(self, x, obj):
        if self.fy >= 0:
            self.real_down(x, obj)
        else:
            self.real_up(x, obj)

    def right(self, x, obj):
        if obj.type == "wall":
            self.set_position(obj.obj.x + obj.obj.width, self.y)

    def up(self, x, obj):
        if self.fy >= 0:
            self.real_down(x, obj)
        else:
            self.real_up(x, obj)

    collide = [left, down, right, up, None]


@Existence.logic_class
class Player:
    def __init__(self):
        self.pictures = [self.picture, pygame.transform.flip(self.picture,
                                                             True, False)]
        self.left = False

        self.keyboard = self.props["keyboard"]
        self.dark = self.props["dark"]
        self.set_light()
        self.event_dealer = self.props["event dealer"]
        self.display = self.props["display"]
        self.live = 500
        self.cold = 0
        self.food = 500

        self.map_width = self.Map.width
        self.map_height = self.Map.height

        self.width = self.obj.width
        self.height = self.obj.height
        self.ground = False
        self.ignore = False
        self.door = None
        self.out = self.props["outside"]

        self.out_collide = []
        self.in_collide = []

    def set_light(self):
        x, y = self.get_blit()
        self.dark.set_pos(x + self.width / 2, y + self.height / 2)

    def blit(self):
        if self.cl_objects and not self.out_collide:
            for obj in self.cl_objects:
                if "in" in obj.opts:
                    self.in_collide.append(obj)
                elif "out" in obj.opts:
                    self.out_collide.append(obj)
                else:
                    self.in_collide.append(obj)
                    self.out_collide.append(obj)
            self.cl_objects = self.out_collide

        if self.keyboard["right"]:
            self.move(5, 0)
            self.left = False
        if self.keyboard["left"]:
            self.move(-5, 0)
            self.left = True
        if self.keyboard["up"] and self.ground:
            self.ground = False
            self.fy = -12
        if self.keyboard["down"] == 1 and self.ground:
            self.ignore = self.ground
            self.ground = False
        if self.keyboard["hit"] == 1 and self.door:
            self.event_dealer.add("go", "in" if self.out else "out")
            self.out = not self.out
            if self.out:
                self.cl_objects = self.out_collide
            else:
                self.cl_objects = self.in_collide
        self.door = None
        
        if self.out:
            if self.cold >= 500:
                self.cold = 500
                self.live -= 0.1
            else:
                self.cold += 0.2
        elif self.cold > 0:
            self.cold -= 0.05
        else:
            self.cold = 0
        if self.food > 0:
            self.food -= 0.05
        else:
            self.food = 0
            self.live -= 0.1

        if self.live <= 0:
            self.event_dealer.add("die")
        self.display.set_width(int(self.live), int(self.cold), int(self.food))

        if self.x < 0:
            self.set_position(0, self.y)
        if self.x + self.width > self.map_width:
            self.set_position(self.map_width - self.width, self.y)
        if self.y > self.map_height:
            self.event_dealer.add("die")

        self.picture = self.pictures[self.left]
        self.set_light()

    def left(self, x, obj):
        if obj.type == "wall":
            self.set_position(obj.obj.x - self.width, self.y)
        if obj.type == "cactus":
            self.live -= 5
            self.set_position(obj.obj.x - self.width, self.y)
            obj.move(10, 0)

    def real_down(self, x, obj):
        if obj.type == "wall":
            self.set_position(self.x, obj.obj.y - self.height)
            self.fy = 0
            self.ground = obj
            self.ignore = False
        if (obj.type == "platform" and obj != self.ignore and
                self.fy > 0.1):
            self.set_position(self.x, obj.obj.y - self.height)
            self.fy = 0
            self.ground = obj
            self.ignore = False
        if obj.type == "cactus":
            self.live -= 5
            self.set_position(self.x, obj.obj.y - self.height)
            self.fy = -5
            obj.fy += 5

    def real_up(self, x, obj):
        if obj.type == "wall":
            self.set_position(self.x, obj.obj.y + obj.obj.height)
            self.fy = 0
        if obj.type == "cactus":
            self.live -= 5
            self.set_position(self.x, obj.obj.y + obj.obj.height)
            self.fy = 0
            obj.fy -= 5

    def down(self, x, obj):
        if self.fy >= 0:
            self.real_down(x, obj)
        else:
            self.real_up(x, obj)

    def right(self, x, obj):
        if obj.type == "wall":
            self.set_position(obj.obj.x + obj.obj.width, self.y)
        if obj.type == "cactus":
            self.live -= 5
            self.set_position(obj.obj.x + obj.obj.width, self.y)
            obj.move(-10, 0)


    def up(self, x, obj):
        if self.fy >= 0:
            self.real_down(x, obj)
        else:
            self.real_up(x, obj)

    def any_(self, obj):
        if obj.type == "door":
            self.door = obj
        if obj.type == "item":
            if obj.available() and self.keyboard["hit"] == 1:
                item = obj.take()
                self.items[item](self)

    collide = [left, down, right, up, any_]

    def bottle(self):
        self.cold -= 100
        self.cold = max(0, self.cold)

    def sandwich(self):
        self.food += 150
        if self.food > 500:
            self.food = 500 - self.food % 500

    items = {"bottle": bottle, "sandwich": sandwich}


class Level:
    layers = {"inside": SwitchLayer, "outside": SwitchLayer}
    shared = {}
    images = {
        "player": pygame.image.load(consts.join(consts.ART_PATH,
                                                "player.png")),
        "cactus": pygame.image.load(consts.join(consts.ART_PATH,
                                                "cactus.png"))
    }
    decode = {
        "item": Existence.mov_type(
            None, None, props=shared, logic=ItemGenerator),
        "player": Existence.mov_type(
            None, images["player"], props=shared, logic=Existence.Subject +
            Existence.opt_size + Existence.gravity(0.5) +
            Existence.collider2("wall", "platform", "door", "item", "cactus") +
            Player),
        "cactus": Existence.mov_type(
            None, images["cactus"], props=shared, logic=Existence.opt_size +
            Existence.collider2("wall", "platform") + Existence.gravity(0.5) +
            Cactus)
    }
    def __init__(self, name, game):
        self.game = game
        self.keyboard = self.shared["keyboard"] = game.keyboard
        self.dark = self.shared["dark"] = game.dark
        self.shared["outside"] = MutableBool(True)
        self.ggame = game.game
        self.main_place = game.places[0]
        self.quit_place = game.places[2]
        self.event_dealer = self.shared["event dealer"] = game.event_dealer
        self.display = self.shared["display"] = Display(game.screen)
        self.stopped = True

        self.name = name
        self.map = Orientation.tiled_map(consts.join(consts.MAP_PATH, name),
                                         [self.layers, {}, self.decode])

        self.inside = []
        self.outside = []
        for layer in self.map.layers:
            if "inside" in layer.name:
                self.inside.append(layer)
                layer.turn_off()
            elif "outside" in layer.name:
                self.outside.append(layer)

    def blit(self):
        if self.stopped:
            self.display.start()
            self.stopped = False

        self.keyboard.update()

        self.map.blit()

        for event, info in self.event_dealer.get():
            if event == "die":
                self.ggame.change(self.quit_place)
            if event == "go":
                if info == "out":
                    self.go_out()
                else:
                    self.go_in()

        self.dark.blit()

        self.display.blit()
        
        if self.keyboard["quit"] == 1:
            self.ggame.change(self.main_place)
            self.keyboard.update()
            self.display.stop()
            self.stopped = True

    def go_in(self):
        self.shared["outside"].set(False)
        for layer in self.inside:
            layer.turn_on()
        for layer in self.outside:
            layer.turn_off()

    def go_out(self):
        self.shared["outside"].set(True)
        for layer in self.inside:
            layer.turn_off()
        for layer in self.outside:
            layer.turn_on()
